use crate::tokens::Position;

#[derive(Debug)]
pub struct Node<T> {
    pub value: T,
    pub pos: Position,
}

#[derive(Debug)]
pub enum Expr {
    Error(String),
    Unit,
    Id(String),
    Int(i64),
    Float(f64),
    String(String),
    Bool(bool),
    Array(Vec<Node<Expr>>),
    Object(Vec<Node<ObjectEntry>>),
    Tuple(Vec<Node<Expr>>),

    If {
        condition: Box<Node<Expr>>,
        truthy: Box<Node<Expr>>,
        falsy: Option<Box<Node<Expr>>>,
    },
    FunctionValue {
        parameters: Vec<Node<Parameter>>,
        body: Box<Node<Expr>>,
    },

    Member {
        subject: Box<Node<Expr>>,
        value: String,
    },
    Index {
        subject: Box<Node<Expr>>,
        value: Box<Node<Expr>>,
    },
    Call {
        subject: Box<Node<Expr>>,
        arguments: Vec<Node<Expr>>,
    },
    Unary {
        unary_type: UnaryType,
        subject: Box<Node<Expr>>,
    },
    Binary {
        binary_type: BinaryType,
        left: Box<Node<Expr>>,
        right: Box<Node<Expr>>,
    },
    RangeExclusive {
        begin: Box<Node<Expr>>,
        end: Box<Node<Expr>>,
    },
    RangeInclusive {
        begin: Box<Node<Expr>>,
        end: Box<Node<Expr>>,
    },

    Assign {
        assign_type: AssignType,
        subject: Box<Node<Expr>>,
        value: Box<Node<Expr>>,
    },
    Let {
        subject: Node<Parameter>,
        value: Option<Box<Node<Expr>>>,
    },
    Continue,
    Break,
    While {
        condition: Box<Node<Expr>>,
        body: Box<Node<Expr>>,
    },
    For {
        subject: Node<Parameter>,
        value: Box<Node<Expr>>,
        body: Box<Node<Expr>>,
    },
    Return(Option<Box<Node<Expr>>>),
    Function {
        name: String,
        parameters: Vec<Node<Parameter>>,
        body: Box<Node<Expr>>,
    },
    Block {
        statements: Vec<Node<Expr>>,
        value: Option<Box<Node<Expr>>>,
    },

    Spread(Box<Node<Expr>>),
}

#[derive(Debug)]
pub enum ObjectEntry {
    Error(String),
    Pair {
        key: Box<Node<Expr>>,
        value: Box<Node<Expr>>,
    },
    Spread(Box<Node<Expr>>),
}

#[derive(Debug)]
pub enum UnaryType {
    Not,
    Negate,
    Reference,
    ReferenceMut,
    Dereference,
}

#[derive(Debug)]
pub enum BinaryType {
    Exponentiate,
    Multiply,
    Divide,
    Modulo,
    Add,
    Subtract,
    LT,
    LTE,
    GT,
    GTE,
    In,
    NotIn,
    Equal,
    Inequal,
    And,
    Or,
}

#[derive(Debug)]
pub enum AssignType {
    Assign,
    Add,
    Subtract,
    Multiply,
    Divide,
    Modulo,
}

#[derive(Debug)]
pub enum Parameter {
    Error(String),
    Id { name: String, mutable: bool },
    Spread(Box<Node<Parameter>>),
}
