use crate::tokens::{Position, PositionKnowing, Token, TokenType};
use std::str::Chars;

pub struct Lexer<'a> {
    text: &'a str,
    chars: Chars<'a>,
    current_char: Option<char>,
    index: usize,
    line: i32,
    col: i32,
}

impl<'a> Lexer<'a> {
    pub fn new(text: &'a str) -> Self {
        let mut chars = text.chars();
        let first_char = chars.next();
        Self {
            text,
            chars,
            current_char: first_char,
            index: 0,
            line: 1,
            col: 1,
        }
    }

    fn next_token(&mut self) -> Option<Token> {
        if self.done() {
            return None;
        }
        match self.current() {
            ' ' | '\t' | '\r' | '\n' => self.skip_whitespace(),
            '1'..='9' => Some(self.int_token()),
            'a'..='z' | 'A'..='Z' | '_' => Some(self.id_token()),
            '"' => Some(self.string_token()),
            '+' => {
                Some(self.single_or_double_char_token(TokenType::Plus, '=', TokenType::PlusEqual))
            }
            '-' => {
                Some(self.single_or_double_char_token(TokenType::Minus, '=', TokenType::MinusEqual))
            }
            '*' => Some(self.asterisk_token()),
            '/' => self.slash_token(),
            '%' => Some(self.single_or_double_char_token(
                TokenType::Percent,
                '=',
                TokenType::PercentEqual,
            )),
            '=' => Some(self.equal_token()),
            '!' => Some(self.single_or_double_char_token(
                TokenType::Exclamation,
                '=',
                TokenType::ExclamationEqual,
            )),
            '<' => Some(self.single_or_double_char_token(
                TokenType::LessThan,
                '=',
                TokenType::LessThanEqual,
            )),
            '>' => Some(self.single_or_double_char_token(
                TokenType::GreaterThan,
                '=',
                TokenType::GreaterThanEqual,
            )),
            '(' => Some(self.step_and_token(TokenType::LParen, self.pos())),
            ')' => Some(self.step_and_token(TokenType::RParen, self.pos())),
            '{' => Some(self.step_and_token(TokenType::LBrace, self.pos())),
            '}' => Some(self.step_and_token(TokenType::RBrace, self.pos())),
            '[' => Some(self.step_and_token(TokenType::LBracket, self.pos())),
            ']' => Some(self.step_and_token(TokenType::RBracket, self.pos())),
            '.' => Some(self.dot_token()),
            ',' => Some(self.step_and_token(TokenType::Comma, self.pos())),
            ':' => Some(self.step_and_token(TokenType::Colon, self.pos())),
            ';' => Some(self.step_and_token(TokenType::Semicolon, self.pos())),
            '&' => Some(self.step_and_token(TokenType::Ampersand, self.pos())),
            _ => Some(self.step_and_token(TokenType::InvalidChar, self.pos())),
        }
    }

    fn skip_whitespace(&mut self) -> Option<Token> {
        while !self.done() && matches!(self.current(), ' ' | '\t' | '\r' | '\n') {
            self.step()
        }
        self.next_token()
    }

    fn int_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        while !self.done() && matches!(self.current(), '0'..='9') {
            self.step();
        }
        self.token(TokenType::Int, start)
    }

    fn string_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        let mut escaped = false;
        while !self.done() && (self.current() != '"' || escaped) {
            escaped = self.current() == '\\' && !escaped;
            self.step();
        }
        if self.done() || self.current() != '"' {
            self.token(TokenType::MalformedString, start)
        } else {
            self.step_and_token(TokenType::String, start)
        }
    }

    fn id_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        while !self.done() && matches!(self.current(), 'a'..='z' | 'A'..='Z' | '0'..='9' | '_') {
            self.step();
        }
        self.token(
            match &self.text[start.index..self.index] {
                "false" => TokenType::False,
                "true" => TokenType::True,
                "let" => TokenType::Let,
                "mut" => TokenType::Mut,
                "if" => TokenType::If,
                "else" => TokenType::Else,
                "while" => TokenType::While,
                "for" => TokenType::For,
                "in" => TokenType::In,
                "break" => TokenType::Break,
                "continue" => TokenType::Continue,
                "fn" => TokenType::Fn,
                "return" => TokenType::Return,
                "end" => TokenType::End,
                "_" => TokenType::Underscore,
                _ => TokenType::Id,
            },
            start,
        )
    }

    fn single_or_double_char_token(
        &mut self,
        single_type: TokenType,
        double_char: char,
        double_type: TokenType,
    ) -> Token {
        let start = self.pos();
        self.step();
        if !self.done() && self.current() == double_char {
            self.step_and_token(double_type, start)
        } else {
            self.token(single_type, start)
        }
    }

    fn asterisk_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        if !self.done() && self.current() == '*' {
            self.step();
            if !self.done() && self.current() == '=' {
                self.step_and_token(TokenType::DoubleAsteriskEqual, start)
            } else {
                self.token(TokenType::DoubleAsterisk, start)
            }
        } else if !self.done() && self.current() == '=' {
            self.step_and_token(TokenType::AsteriskEqual, start)
        } else {
            self.token(TokenType::Asterisk, start)
        }
    }

    fn slash_token(&mut self) -> Option<Token> {
        let start = self.pos();
        self.step();
        if !self.done() && self.current() == '/' {
            self.step();
            while !self.done() && self.current() != '\n' {
                self.step();
            }
            self.next_token()
        } else if !self.done() && self.current() == '*' {
            self.step();
            let mut depth = 1;
            let mut last_char: Option<char> = None;
            while !self.done() {
                match (last_char, self.current()) {
                    (Some('/'), '*') => {
                        depth += 1;
                    }
                    (Some('*'), '/') => {
                        depth -= 1;
                        if depth == 0 {
                            self.step();
                            break;
                        }
                    }
                    _ => {}
                }
                last_char = Some(self.current());
                self.step();
            }
            if depth != 0 {
                Some(self.token(TokenType::MalformedComment, start))
            } else {
                self.next_token()
            }
        } else if !self.done() && self.current() == '=' {
            return Some(self.step_and_token(TokenType::SlashEqual, start));
        } else {
            return Some(self.token(TokenType::Slash, start));
        }
    }

    fn equal_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        if !self.done() && self.current() == '=' {
            self.step_and_token(TokenType::EqualEqual, start)
        } else if !self.done() && self.current() == '>' {
            self.step_and_token(TokenType::EqualLessThan, start)
        } else {
            self.step_and_token(TokenType::Equal, start)
        }
    }

    fn dot_token(&mut self) -> Token {
        let start = self.pos();
        self.step();
        if !self.done() && matches!(self.current(), '0'..='9') {
            self.step();
            while !self.done() && matches!(self.current(), '0'..='9') {
                self.step();
            }
            self.token(TokenType::Decimal, start)
        } else if !self.done() && self.current() == '.' {
            self.step();
            if !self.done() && self.current() == '.' {
                self.step_and_token(TokenType::DotDotDot, start)
            } else if !self.done() && self.current() == '=' {
                self.step_and_token(TokenType::DotDotEqual, start)
            } else {
                self.token(TokenType::DotDot, start)
            }
        } else {
            self.token(TokenType::Dot, start)
        }
    }

    fn step_and_token(&mut self, token_type: TokenType, start: Position) -> Token {
        self.step();
        self.token(token_type, start)
    }

    fn token(&self, token_type: TokenType, start: Position) -> Token {
        Token {
            token_type,
            length: self.index - start.index,
            pos: start,
        }
    }

    fn done(&self) -> bool {
        self.current_char.is_none()
    }

    fn current(&self) -> char {
        self.current_char.expect("done() checked")
    }

    fn step(&mut self) {
        self.index += 1;
        if !self.done() {
            if self.current() == '\n' {
                self.line += 1;
                self.col = 1;
            } else {
                self.col += 1;
            }
        }
        self.current_char = self.chars.next();
    }
}

impl<'a> Iterator for Lexer<'a> {
    type Item = Token;

    fn next(&mut self) -> Option<Self::Item> {
        self.next_token()
    }
}

impl<'a> PositionKnowing for Lexer<'a> {
    fn pos(&self) -> Position {
        Position {
            index: self.index,
            line: self.line,
            col: self.col,
        }
    }
}
