use std::iter::Peekable;
use std::str::Chars;

use crate::ast::{AssignType, BinaryType, Expr, Node, ObjectEntry, Parameter, UnaryType};
use crate::tokens::{PositionKnowing, Token, TokenType};

pub struct Parser<'a, Tokens>
where
    Tokens: PositionKnowing + Iterator<Item = Token>,
{
    text: &'a str,
    tokens: Tokens,
    current_token: Option<Token>,
}

impl<'a, Tokens> Parser<'a, Tokens>
where
    Tokens: PositionKnowing + Iterator<Item = Token>,
{
    pub fn new(text: &'a str, mut lexer: Tokens) -> Self {
        Self {
            text,
            current_token: lexer.next(),
            tokens: lexer,
        }
    }

    pub fn parse_top_level_statements(&mut self) -> Vec<Node<Expr>> {
        let mut statements = Vec::<Node<Expr>>::new();
        while !self.done() {
            statements.push(self.parse_statement());
        }
        statements
    }

    fn parse_block(&mut self) -> Node<Expr> {
        self.step();
        let mut statements = Vec::<Node<Expr>>::new();
        let mut last = None;
        while !self.done() && !self.current_is(TokenType::RBrace) {
            if let Some(statement) = last {
                statements.push(statement);
            }
            last = Some(self.parse_statement());
        }
        if self.done() || !self.current_is(TokenType::RBrace) {
            statements.push(self.error("expected '}'"));
        }
        self.node(Expr::Block {
            statements,
            value: last.map(|v| Box::new(v)),
        })
    }

    pub fn parse_statement(&mut self) -> Node<Expr> {
        if self.done() {
            return self.error("expected statement");
        }
        match self.current().token_type {
            TokenType::Fn => self.parse_function(),
            TokenType::Return => self.parse_return(),
            TokenType::For => self.parse_for(),
            TokenType::While => self.parse_while(),
            TokenType::Break => self.parse_break(),
            TokenType::Continue => self.parse_continue(),
            TokenType::Let => self.parse_let(),
            _ => self.parse_assign(),
        }
    }

    fn parse_function(&mut self) -> Node<Expr> {
        self.step();
        if self.done() || !self.current_is(TokenType::Id) {
            return self.error("expected Id");
        }
        let name = self.token_string(self.current());
        self.step();
        let (parameters, body) = match self.parse_function_details() {
            Ok(v) => v,
            Err(e) => return e,
        };
        self.node(Expr::Function {
            name,
            parameters,
            body: Box::new(body),
        })
    }

    fn parse_function_details(&mut self) -> Result<(Vec<Node<Parameter>>, Node<Expr>), Node<Expr>> {
        if self.done() || !self.current_is(TokenType::LParen) {
            return Err(self.error("expected '('"));
        }
        self.step();
        let mut parameters = self.parse_function_parameters();
        if self.done() || !self.current_is(TokenType::RParen) {
            return Err(self.error("expected ')'"));
        }
        self.step();
        let body = if !self.done() && self.current_is(TokenType::LBrace) {
            self.parse_block()
        } else if !self.done() && self.current_is(TokenType::EqualLessThan) {
            self.step();
            self.parse_expr()
        } else {
            return Err(self.error("expected '{'"));
        };
        Ok((parameters, body))
    }

    fn parse_function_parameters(&mut self) -> Vec<Node<Parameter>> {
        let mut parameters = Vec::<Node<Parameter>>::new();
        if !self.done() && !self.current_is(TokenType::RParen) {
            if self.current_is(TokenType::DotDotDot) {
                self.step();
                let inner = self.parse_parameter();
                parameters.push(self.node(Parameter::Spread(Box::new(inner))));
            } else {
                parameters.push(self.parse_parameter());
                while !self.done() && self.current_is(TokenType::Comma) {
                    self.step();
                    if self.current_is(TokenType::DotDotDot) {
                        self.step();
                        let inner = self.parse_parameter();
                        parameters.push(self.node(Parameter::Spread(Box::new(inner))));
                        break;
                    }
                    if self.done() || self.current_is(TokenType::RParen) {
                        break;
                    }
                    parameters.push(self.parse_parameter());
                }
            }
        }
        parameters
    }

    fn parse_parameter(&mut self) -> Node<Parameter> {
        let mutable = if !self.done() && self.current_is(TokenType::Mut) {
            self.step();
            true
        } else {
            false
        };
        if self.done() || !self.current_is(TokenType::Id) {
            return self.node(Parameter::Error("expected Id".to_string()));
        }
        let name = self.token_string(self.current());
        self.step_and_node(Parameter::Id { mutable, name })
    }

    fn parse_return(&mut self) -> Node<Expr> {
        self.step();
        if !self.done() && !self.current_is(TokenType::Semicolon) {
            let value = self.parse_expr();
            self.node(Expr::Return(Some(Box::new(value))))
        } else {
            self.node(Expr::Return(None))
        }
    }

    fn parse_for(&mut self) -> Node<Expr> {
        self.step();
        let subject = self.parse_parameter();
        if self.done() || !self.current_is(TokenType::In) {
            return self.error("expected 'in'");
        }
        self.step();
        let value = self.parse_expr();
        if self.done() || !self.current_is(TokenType::LBrace) {
            return self.error("expected '{'");
        }
        let body = self.parse_block();
        self.node(Expr::For {
            subject,
            value: Box::new(value),
            body: Box::new(body),
        })
    }

    fn parse_while(&mut self) -> Node<Expr> {
        self.step();
        let condition = self.parse_expr();
        if self.done() || !self.current_is(TokenType::LBrace) {
            return self.error("expected '{'");
        }
        let body = self.parse_block();
        self.node(Expr::While {
            condition: Box::new(condition),
            body: Box::new(body),
        })
    }

    fn parse_break(&mut self) -> Node<Expr> {
        self.step_and_node(Expr::Continue)
    }

    fn parse_continue(&mut self) -> Node<Expr> {
        self.step_and_node(Expr::Continue)
    }

    fn parse_let(&mut self) -> Node<Expr> {
        self.step();
        let subject = self.parse_parameter();
        if !self.done() && self.current_is(TokenType::Equal) {
            self.step();
            let value = self.parse_expr();
            self.node(Expr::Let {
                subject,
                value: Some(Box::new(value)),
            })
        } else {
            self.node(Expr::Let {
                subject,
                value: None,
            })
        }
    }

    fn parse_assign(&mut self) -> Node<Expr> {
        let subject = self.parse_expr();
        if self.done() {
            return subject;
        }
        match self.current().token_type {
            TokenType::Equal => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Assign,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            TokenType::PlusEqual => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Add,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            TokenType::MinusEqual => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Subtract,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            TokenType::AsteriskEqual => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Multiply,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            TokenType::SlashEqual => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Divide,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            TokenType::PercentEqual => {
                self.step();
                let value = self.parse_expr();
                self.node(Expr::Assign {
                    assign_type: AssignType::Modulo,
                    subject: Box::new(subject),
                    value: Box::new(value),
                })
            }
            _ => subject,
        }
    }

    fn parse_spread(&mut self) -> Node<Expr> {
        self.step();
        let inner = self.parse_expr();
        self.node(Expr::Spread(Box::new(inner)))
    }

    pub fn parse_expr(&mut self) -> Node<Expr> {
        self.parse_prec_or()
    }

    fn parse_prec_or(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_and();
        while !self.done() {
            if self.current_is(TokenType::Or) {
                self.step();
                let right = self.parse_prec_and();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Or,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_and(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_equal_inequal();
        while !self.done() {
            if self.current_is(TokenType::And) {
                self.step();
                let right = self.parse_prec_equal_inequal();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::And,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_equal_inequal(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_lt_lte_gt_gte_in();
        while !self.done() {
            if self.current_is(TokenType::EqualEqual) {
                self.step();
                let right = self.parse_prec_lt_lte_gt_gte_in();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Equal,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::ExclamationEqual) {
                self.step();
                let right = self.parse_prec_lt_lte_gt_gte_in();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Inequal,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_lt_lte_gt_gte_in(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_range();
        while !self.done() {
            if self.current_is(TokenType::LessThan) {
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::LT,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::GreaterThan) {
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::GT,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::LessThanEqual) {
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::LTE,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::GreaterThanEqual) {
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::GTE,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::In) {
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::In,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Not) {
                self.step();
                if self.done() || !self.current_is(TokenType::In) {
                    return self.error("expected 'in'");
                }
                self.step();
                let right = self.parse_prec_range();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::NotIn,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_range(&mut self) -> Node<Expr> {
        let begin = self.parse_prec_add_subtract();
        if self.done() {
            return begin;
        }
        match self.current().token_type {
            TokenType::DotDot => {
                self.step();
                let end = self.parse_prec_add_subtract();
                self.node(Expr::RangeExclusive {
                    begin: Box::new(begin),
                    end: Box::new(end),
                })
            }
            TokenType::DotDotEqual => {
                self.step();
                let end = self.parse_prec_add_subtract();
                self.node(Expr::RangeExclusive {
                    begin: Box::new(begin),
                    end: Box::new(end),
                })
            }
            _ => begin,
        }
    }

    fn parse_prec_add_subtract(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_multiply_divide_modulo();
        while !self.done() {
            if self.current_is(TokenType::Plus) {
                self.step();
                let right = self.parse_prec_multiply_divide_modulo();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Add,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Minus) {
                self.step();
                let right = self.parse_prec_multiply_divide_modulo();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Subtract,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_multiply_divide_modulo(&mut self) -> Node<Expr> {
        let mut left = self.parse_prec_unary();
        while !self.done() {
            if self.current_is(TokenType::Asterisk) {
                self.step();
                let right = self.parse_prec_unary();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Multiply,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Slash) {
                self.step();
                let right = self.parse_prec_unary();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Divide,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Percent) {
                self.step();
                let right = self.parse_prec_unary();
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Modulo,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        left
    }

    fn parse_prec_unary(&mut self) -> Node<Expr> {
        if !self.done() && self.current_is(TokenType::Not) {
            self.step();
            let subject = Box::new(self.parse_prec_unary());
            self.node(Expr::Unary {
                unary_type: UnaryType::Not,
                subject,
            })
        } else if !self.done() && self.current_is(TokenType::Minus) {
            self.step();
            let subject = Box::new(self.parse_prec_unary());
            self.node(Expr::Unary {
                unary_type: UnaryType::Negate,
                subject,
            })
        } else if !self.done() && self.current_is(TokenType::Ampersand) {
            self.step();
            if !self.done() && self.current_is(TokenType::Mut) {
                self.step();
                let subject = Box::new(self.parse_prec_unary());
                self.node(Expr::Unary {
                    unary_type: UnaryType::ReferenceMut,
                    subject,
                })
            } else {
                let subject = Box::new(self.parse_prec_unary());
                self.node(Expr::Unary {
                    unary_type: UnaryType::Reference,
                    subject,
                })
            }
        } else if !self.done() && self.current_is(TokenType::Asterisk) {
            self.step();
            let subject = Box::new(self.parse_prec_unary());
            self.node(Expr::Unary {
                unary_type: UnaryType::Dereference,
                subject,
            })
        } else {
            self.parse_prec_exponentiate()
        }
    }

    fn parse_prec_exponentiate(&mut self) -> Node<Expr> {
        let left = self.parse_prec_member_index_call();
        if !self.done() && self.current_is(TokenType::AsteriskEqual) {
            let right = self.parse_prec_exponentiate();
            self.step_and_node(Expr::Binary {
                binary_type: BinaryType::Exponentiate,
                left: Box::new(left),
                right: Box::new(right),
            })
        } else {
            left
        }
    }

    fn parse_prec_member_index_call(&mut self) -> Node<Expr> {
        let mut subject = self.parse_operand();
        while !self.done() {
            if self.current_is(TokenType::Dot) {
                self.step();
                if self.done() || self.current_is(TokenType::Id) {
                    return self.error("expected identifier");
                }
                let value = self.token_string(self.current());
                self.step();
                subject = self.node(Expr::Member {
                    subject: Box::new(subject),
                    value,
                });
            } else if self.current_is(TokenType::LBracket) {
                self.step();
                let value = self.parse_expr();
                if self.done() || !self.current_is(TokenType::RBracket) {
                    return self.error("expected ']'");
                }
                subject = self.node(Expr::Index {
                    subject: Box::new(subject),
                    value: Box::new(value),
                });
            } else if self.current_is(TokenType::LParen) {
                self.step();
                let mut arguments = Vec::<Node<Expr>>::new();
                if !self.done() && !self.current_is(TokenType::RParen) {
                    arguments.push(self.parse_expr());
                    while !self.done() && self.current_is(TokenType::Comma) {
                        self.step();
                        if self.done() || self.current_is(TokenType::RParen) {
                            self.step();
                        }
                        arguments.push(self.parse_expr());
                    }
                }
                if self.done() || !self.current_is(TokenType::RParen) {
                    return self.error("expected ')'");
                }
                self.step();
                subject = self.node(Expr::Call {
                    subject: Box::new(subject),
                    arguments,
                })
            } else {
                break;
            }
        }
        subject
    }

    fn parse_operand(&mut self) -> Node<Expr> {
        if self.done() {
            return self.error("expected value, got eof");
        }
        match &self.current().token_type {
            TokenType::Id => self.step_and_node(Expr::Id(self.token_string(self.current()))),
            TokenType::Int => {
                let mut value_string = self.token_string(self.current());
                self.step();
                if !self.done() && self.current_is(TokenType::Decimal) {
                    value_string.push_str(&self.token_string(self.current()));
                    self.step_and_node(Expr::Float(value_string.parse::<f64>().expect("valid f64")))
                } else {
                    self.node(Expr::Int((value_string).parse::<i64>().expect("valid i64")))
                }
            }
            TokenType::Decimal => self.step_and_node(Expr::Float(
                self.token_string(self.current())
                    .parse::<f64>()
                    .expect("valid f64"),
            )),
            TokenType::String => self.parse_string(),
            TokenType::False => self.step_and_node(Expr::Bool(false)),
            TokenType::True => self.step_and_node(Expr::Bool(true)),
            TokenType::LParen => self.parse_unit_group_or_tuple(),
            TokenType::LBrace => self.parse_block(),
            TokenType::Underscore => self.parse_object(),
            TokenType::LBracket => self.parse_array(),
            TokenType::If => self.parse_if(),
            TokenType::Fn => self.parse_function_value(),
            TokenType::MalformedString => self.error("malformed string"),
            other => self.node(Expr::Error(format!("expected value, got {:?}", other))),
        }
    }

    fn parse_string(&mut self) -> Node<Expr> {
        let value_result = parse_string_value(&self.token_string(self.current()));
        self.step_and_node(match value_result {
            Ok(value) => Expr::String(value),
            Err(error) => Expr::Error(error),
        })
    }

    fn parse_unit_group_or_tuple(&mut self) -> Node<Expr> {
        self.step();
        if !self.done() && !self.current_is(TokenType::LParen) {
            let first_value = self.parse_expr();
            if !self.done() && !self.current_is(TokenType::RParen) {
                let mut values = vec![first_value];
                while !self.done() && self.current_is(TokenType::Comma) {
                    self.step();
                    if self.done() || self.current_is(TokenType::RParen) {
                        break;
                    }
                    values.push(self.parse_expr());
                }
                if self.done() || !self.current_is(TokenType::RParen) {
                    return self.error("expected ')'");
                }
                self.step_and_node(Expr::Tuple(values))
            } else {
                self.step_and(first_value)
            }
        } else {
            self.step_and_node(Expr::Unit)
        }
    }

    fn parse_object(&mut self) -> Node<Expr> {
        self.step();
        if self.done() || !self.current_is(TokenType::LBrace) {
            return self.error("expected '{' of object literal");
        }
        self.step();
        let mut values = Vec::<Node<ObjectEntry>>::new();
        if !self.done() && !self.current_is(TokenType::RBracket) {
            if self.current_is(TokenType::DotDotDot) {
                values.push(self.parse_object_entry());
            } else {
                values.push(self.parse_object_entry());
                while !self.done() && self.current_is(TokenType::Comma) {
                    self.step();
                    if self.done() || self.current_is(TokenType::RBracket) {
                        break;
                    }
                    if self.current_is(TokenType::DotDotDot) {
                        values.push(self.parse_object_entry());
                        break;
                    }
                    values.push(self.parse_object_entry());
                }
            }
        }
        if self.done() || !self.current_is(TokenType::RBracket) {
            return self.error("expected ']'");
        }
        self.step_and_node(Expr::Object(values))
    }

    fn parse_object_entry(&mut self) -> Node<ObjectEntry> {
        if self.done() {
            return self.node(ObjectEntry::Error("expected object entry".to_string()));
        }
        match self.current().token_type {
            TokenType::DotDotDot => {
                let expr = self.parse_spread();
                self.node(ObjectEntry::Spread(Box::new(expr)))
            }
            TokenType::String => {
                let key = self.parse_string();
                if self.done() || !self.current_is(TokenType::Colon) {
                    return self.node(ObjectEntry::Error("expected ':'".to_string()));
                }
                self.step();
                let value = self.parse_expr();
                self.node(ObjectEntry::Pair {
                    key: Box::new(key),
                    value: Box::new(value),
                })
            }
            _ => self.node(ObjectEntry::Error("expected object entry".to_string())),
        }
    }

    fn parse_array(&mut self) -> Node<Expr> {
        self.step();
        let mut values = Vec::<Node<Expr>>::new();
        if !self.done() && !self.current_is(TokenType::RBracket) {
            if self.current_is(TokenType::DotDotDot) {
                values.push(self.parse_spread());
            } else {
                values.push(self.parse_expr());
                while !self.done() && self.current_is(TokenType::Comma) {
                    self.step();
                    if self.done() || self.current_is(TokenType::RBracket) {
                        break;
                    }
                    if self.current_is(TokenType::DotDotDot) {
                        values.push(self.parse_spread());
                        break;
                    }
                    values.push(self.parse_expr());
                }
            }
        }
        if self.done() || !self.current_is(TokenType::RBracket) {
            return self.error("expected ']'");
        }
        self.step_and_node(Expr::Array(values))
    }

    fn parse_if(&mut self) -> Node<Expr> {
        self.step();
        let condition = self.parse_expr();
        if self.done() || !self.current_is(TokenType::LBrace) {
            return self.error("expected '{'");
        }
        let truthy = self.parse_block();
        let mut falsy = None;
        if !self.done() && self.current_is(TokenType::Else) {
            self.step();
            if self.done() || !self.current_is(TokenType::LBrace) {
                return self.error("expected '{'");
            }
            falsy = Some(self.parse_block());
        }
        self.node(Expr::If {
            condition: Box::new(condition),
            truthy: Box::new(truthy),
            falsy: falsy.map(Box::new),
        })
    }

    fn parse_function_value(&mut self) -> Node<Expr> {
        self.step();
        let (parameters, body) = match self.parse_function_details() {
            Ok(v) => v,
            Err(e) => return e,
        };
        self.node(Expr::FunctionValue {
            parameters,
            body: Box::new(body),
        })
    }

    fn token_string(&self, token: &Token) -> String {
        self.text[token.pos.index..token.pos.index + token.length].to_string()
    }

    fn step_and_node<T>(&mut self, value: T) -> Node<T> {
        self.step_and(self.node(value))
    }

    fn node<T>(&self, value: T) -> Node<T> {
        Node {
            value,
            pos: self.tokens.pos(),
        }
    }

    fn step_and<T>(&mut self, value: T) -> T {
        self.step();
        value
    }

    fn error(&self, message: &str) -> Node<Expr> {
        self.node(Expr::Error(message.to_string()))
    }

    fn done(&self) -> bool {
        self.current_token.is_none()
    }

    fn current_is(&self, token_type: TokenType) -> bool {
        self.current().token_type == token_type
    }

    fn current(&self) -> &Token {
        self.current_token.as_ref().expect("done() checked")
    }

    fn step(&mut self) {
        self.current_token = self.tokens.next();
    }
}

fn parse_string_value(text: &str) -> Result<String, String> {
    let mut input = text.chars().peekable();
    let mut value = String::new();
    input.next().ok_or_else(|| "invalid string".to_string())?;
    while !matches!(input.peek(), Some('"')) {
        value.push(unescape_char_value(&mut input)?)
    }
    Ok(value)
}

fn unescape_char_value(input: &mut Peekable<Chars>) -> Result<char, String> {
    let step_and = |input2: &mut Peekable<Chars>, value: char| {
        input2.next().ok_or("invalid string".to_string())?;
        Ok(value)
    };
    if matches!(input.peek(), Some('\\')) {
        input.next().ok_or("invalid string".to_string())?;
        match input.peek() {
            Some('t') => step_and(input, '\t'),
            Some('r') => step_and(input, '\r'),
            Some('n') => step_and(input, '\n'),
            Some('0') => step_and(input, '\0'),
            Some('x') => {
                input.next().ok_or("invalid string".to_string())?;
                let mut value = String::new();
                for _ in 0..2 {
                    value.push(
                        input
                            .next()
                            .ok_or_else(|| "invalid escape sequence".to_string())?,
                    );
                }
                let int_value = value
                    .parse::<u8>()
                    .map_err(|error| format!("invalid char code: {}", error.to_string()))?;
                char::from_u32(int_value.into()).ok_or_else(|| "invalid char code".to_string())
            }
            Some(_) => input.next().ok_or_else(|| "invalid string".to_string()),
            None => Err("invalid escape sequence".to_string()),
        }
    } else {
        input.next().ok_or("invalid string".to_string())
    }
}
